#include <stdio.h>
#include <stdlib.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/soundcard.h>

#ifdef ENABLE_NLS
#include <locale.h>
#endif

#ifdef USE_ESD
#include <esd.h>

static unsigned char *esd_host =NULL; // NULL = localhost
#endif

#include "../mp3_config.h"
#include "../mp3.h"
#include "../mpg123engine.h"

#include "player.h"
#include "playlist.h"
#include "getlopt.h"

static void usage(char *);
static void print_title(void);

struct parameter param;
static char *listname = NULL;

static int single=0;
static int vol=100,vol_l=100,vol_r=100;

/* Please note: GLO_NUM expects point to LONG! */
topt opts[] = {
//    {'k', "skip",        GLO_ARG | GLO_LONG, 0, &param.startFrame, 0},
//    {'a', "audiodevice", GLO_ARG | GLO_CHAR, 0, &ai.device,  0},
//    {'2', "2to1",        0,                  0, &param.down_sample, 1},
//    {'4', "4to1",        0,                  0, &param.down_sample, 2},
    {'t', "test",        0,                  0, &param.outmode, DECODE_TEST},
//    {'s', "stdout",      0,       SetOutStdout, &param.outmode, DECODE_FILE},
//    {'S', "STDOUT",      0,       SetOutStdout1, &param.outmode, DECODE_AUDIOFILE},
//    {'O', "outfile",     GLO_ARG | GLO_CHAR, SetOutFile, NULL, 0},
///    {'v', "verbose",     0,        set_verbose, 0,           0},
//    {'y', "resync",      0,                  0, &param.tryresync,  FALSE},
#ifdef USE_FAKE_MONO
    {'1',   "left",        0,                  0, &param.force_mono, 1},
    {'2',   "right",       0,                  0, &param.force_mono, 2},
#endif
    {0,   "stereo",      0,                  0, &param.force_stereo, 1},
//    {'g', "gain",        GLO_ARG | GLO_LONG, 0, &ai.gain,    0},
    {'r', "rate",        GLO_ARG | GLO_LONG, 0, &param.force_rate,  0},
    {0,   "8bit",        0,                  0, &param.force_8bit, 1},
//    {'f', "scale",       GLO_ARG | GLO_LONG, 0, &param.outscale,   0},
//    {'n', "frames",      GLO_ARG | GLO_LONG, 0, &numframes,  0},
//    {'b', "buffer",      GLO_ARG | GLO_LONG, 0, &param.usebuffer,  0},
//    {'R', "remote",      0,                  0, &param.remote,     FRONTEND_GENERIC},
    {'R', "remote",      0,                  0, &param.remote,     3},
    {'@', "list",        GLO_ARG | GLO_CHAR, 0, &listname,   0},
    /* 'z' comes from the the german word 'zufall' (eng: random) */
    {'z', "shuffle",     0,                  0, &param.shuffle,    1},
    {'Z', "random",      0,                  0, &param.shuffle,    2},
//    {0,   "aggressive",	 0,   	             0, &param.aggressive,2},
    {0,   "force-3dnowex", 0,                0, &param.simd_select,MODE_3DNOWEXT},
    {0,   "force-3dnow", 0,                  0, &param.simd_select,MODE_3DNOW},
    {0,   "force-sse", 0,                  0, &param.simd_select,MODE_SSE},
    {0,   "force-mmx",   0,                  0, &param.simd_select,MODE_MMX},
    {0,   "force-fpu",   0,                  0, &param.simd_select,MODE_FPU},
    {0,   "test-simd",   0,                  0, &param.test_simd,TRUE},
//#if defined(SET_RT)
//    {'T', "realtime",   0,                  0, &param.realtime, TRUE },
//#else
//    {'T', "realtime",   0,       not_compiled, 0,           0 },    
//#endif
//    {'w', "wav",        GLO_ARG | GLO_CHAR, set_wav, 0 , 0 },
//    {0  , "cdr",        GLO_ARG | GLO_CHAR, set_cdr, 0 , 0 },
//    {0  , "au",         GLO_ARG | GLO_CHAR, set_au,  0 , 0 },
#ifdef USE_ESD
    {0,   "esd",   0,                  0, &param.audio_out,TRUE},
    {0,   "oss",   0,                  0, &param.audio_out,FALSE},
#endif
    {'?', "help",       0,        usage, 0,           0 },
    {'v', "version",    0,                     0, &param.print_version,1},
    {0, 0, 0, 0, 0, 0}
};


#define BUFFLEN 18432
static unsigned char *buff;
static struct AudioInfo_s AudioInfo;

static int OpenFile(char *fname)
{
   int r, mix;
#ifdef USE_ESD
   int esd, esd_rate;
   char pname[32];
   
   AudioInfo.fd = -1;
   esd_format_t format =0;
   esd_info_t *esdctl;
   snprintf(pname, 32, "%s - %d", PACKAGE, getpid());
#endif

   memset(&AudioInfo,0,sizeof (struct AudioInfo_s));
   
DBGMSG0("mp3_open0\n");   
   MP3_Open(fname, BUFFLEN);
   if(MP3Info.filesize==0){  printf("file not found : %s\n", fname);  return 1; }

   MP3_PrintTAG();
   
   if (MP3Info.channels == 2)
     single = -1;
   else
     single = 3;
   
   AudioInfo.fname=fname;
   
//DBGMSG0("fpos=%d, frames=%d, bitrate=%d\n", MP3Info.fpos, MP3Info.frames,MP3Info.bitrate);
   
#ifdef USE_ESD
   if (param.audio_out) 
     {
	if (MP3Info.channels == 2) 
	  format = ESD_BITS16 | ESD_STEREO | ESD_STREAM | ESD_PLAY;
	else
	  format = ESD_BITS16 | ESD_MONO | ESD_STREAM | ESD_PLAY;
     
	esd = esd_open_sound(esd_host);
	AudioInfo.fd = esd_play_stream_fallback(format, MP3Info.samplerate,
						esd_host, pname);
	if (esd >= 0) {
	   DBGMSG0("esd >= 0\n");
	   esdctl = esd_get_all_info(esd);
	   //      esd_print_all_info(esdctl);
	   esd_set_stream_pan(esd, esdctl->player_list->source_id, 
			      vol_l*ESD_VOLUME_BASE/100,vol_r*ESD_VOLUME_BASE/100);
	   esd_rate = esdctl->player_list->rate;
	} else {
	   esd_rate = esd_audio_rate;
	}
	DBGMSG1(" esd_setup: using %dHz samplerate (requested: %d)\n\n",esd_rate,
		MP3Info.samplerate);
	esd_close(esd);
     } else  {
#endif
	AudioInfo.fd=open("/dev/dsp", O_WRONLY); // open audio device
	if(AudioInfo.fd<0){  printf("Can't open audio device\n"); return 1; }
	r=AFMT_S16_LE;ioctl (AudioInfo.fd, SNDCTL_DSP_SETFMT, &r);
	r=MP3Info.channels-1;ioctl (AudioInfo.fd, SNDCTL_DSP_STEREO, &r);
	r=MP3Info.samplerate;ioctl (AudioInfo.fd, SNDCTL_DSP_SPEED, &r);
	//   mix = open("/dev/mixer", O_WRONLY);
	//   r=vol;ioctl(mix,MIXER_WRITE(SOUND_MIXER_PCM),&r);
   
	DBGMSG1(" audio_setup: using %dHz samplerate (requested: %d)\n\n",r,
		MP3Info.samplerate);
#ifdef USE_ESD
     }
#endif	
   
   return 0;
}

static void CloseFile()
{
   close(AudioInfo.fd); // close audio device
   MP3_Close();
   
   sleep(1);
}

static void PlayFile()
{
   int r,len=0,t;
   long playtime;
   char *dirname, *filename;
   
   split_dir_file(AudioInfo.fname, &dirname, &filename);
   printf("Directory: %s\n", dirname);
   printf("Playing MPEG stream from %s\n\n", filename);
   MP3_PrintHeader(); printf("\n");
   AudioInfo.start_time = GetTimer();
   
   if (param.outmode != DECODE_TEST) {
// --- normal decoding mode      
      while(1) {
//      while(AudioInfo.total<10000){
	 int len2;
	 if(len==0) len=MP3_DecodeFrame(buff,single);
	 if(len<=0) break; // EOF
      
//DBGMSG0("fpos=%d, frames=%d, bitrate=%d\n", MP3Info.fpos, MP3Info.frames,MP3Info.bitrate);
	 // play it
	 len2=write(AudioInfo.fd,buff,len);
	 if(len2<0) break; // ERROR?
	 len-=len2; AudioInfo.total+=len2;
	 if(len>0){
	    // this shouldn't happen...
	    memcpy(buff,buff+len2,len);
	    putchar('!');fflush(stdout);
	 }
      }
   } else {
// test mode       
//	while((len=MP3_DecodeFrame(buff,single))>0 && AudioInfo.total<2000000){
      while((len=MP3_DecodeFrame(buff,single))>0){
	 AudioInfo.total+=len;
   
//DBGMSG0("fpos=%d, frames=%d, bitrate=%d\n", MP3Info.fpos, MP3Info.frames,MP3Info.bitrate);
	 // play it
	 // /* DUMP_PCM */ fwrite(audiobuf,len,1,f);
	 //putchar('.');fflush(stdout);
      }
   }

   // AudioInfo.length (play time) = MP3Info.length
   AudioInfo.length = (float)AudioInfo.total/(float)(MP3Info.samplerate*MP3Info.channels*2);
   t = (int)AudioInfo.length/1000;
   printf("%f, %d\n", AudioInfo.length, t);
   printf("[%d:%02d] Decoding of %s finished.\n", t/60, t%60, AudioInfo.fname);
   
   if (param.outmode == DECODE_TEST) {
      playtime = GetTimer() - AudioInfo.start_time;
      printf("\nDecoding time: %8.6f\n",(float)playtime*0.000001f);
      printf("Uncompressed size: %d bytes  (%8.3f secs)\n",AudioInfo.total,AudioInfo.length);
      printf("CPU usage at normal playback: %5.2f %\n",playtime*0.0001f/AudioInfo.length);
//      printf("Uncompressed size: %d bytes  (%8.3f secs)\n",AudioInfo.total,MP3Info.length/1000);
//      printf("CPU usage at normal playback: %5.2f %\n",(float)(playtime*0.1f/MP3Info.length));
   }
}

int main(int argc,char* argv[]){
   int result;
   struct playlist *playlist;
   param.simd_select = MODE_AUTO;
   param.outmode = DECODE_AUDIO;
   param.force_mono = 0;
   param.volume = 127;
   char *fname, *dummy;
   
#ifndef USE_ESD
   param.audio_out = TRUE; /* default is esound */
#endif
#ifdef ENABLE_NLS
   setlocale(LC_ALL, "");
   bindtextdomain(PACKAGE,LOCALEDIR);
   textdomain(PACKAGE);
#endif
   
   if ((buff=(unsigned char *)malloc(sizeof(unsigned char *)*BUFFLEN)) == NULL) {
     printf("%s : not enough memory, cannot allocate.\n");
     exit (1);
   }
   
   while ( (result = getlopt(argc, argv, opts)) != GLO_END) {
      switch(result)
	{
	 case GLO_UNKNOWN:
	   exit (1);
	 case GLO_NOARG:
	   exit (1);
	}
   }
  if (param.print_version) 
     {
	fprintf(stderr,"Version %s\n", VERSION);
	exit (0);
     }
		
   print_title();
   
//   DBGMSG0("simd_select=%d, shuffule=%d\n",param.simd_select,param.shuffle);

   // MPEG Audio init:
#ifdef USE_FAKE_MONO
  MP3_Init(param.force_mono);
  param.auto_simd = modcap.decoder;
  if (param.simd_select != MODE_AUTO) MP3_set_synth_functions(param.simd_select,param.force_mono);
#else
  MP3_Init();
  param.auto_simd = modcap.decoder;
  if (param.simd_select != MODE_AUTO) MP3_set_synth_functions(param.simd_select);
#endif
   
   playlist = new_playlist(argc, argv, listname, loptind);
   if ((!playlist) || (playlist->shuffle_listsize == 0) ) 
     {
	usage(dummy);
	exit (1);
     }
   
   // begin of Loop Playing files
   while (fname = playlist->next(playlist)) 
     {
	if (!*fname || !strcmp(fname, "-"))
	  fname = NULL;
	else
	  {
	     if (OpenFile(fname)==0) 
	       {
		  PlayFile();
		  CloseFile();
	       }
	  }
     }

   free(buff);
}


int split_dir_file (const char *path, char **dname, char **fname)
{
   static char *lastdir = NULL;
   char *slashpos;

   if ((slashpos = strrchr(path, '/'))) {
      *fname = slashpos + 1;
      *dname = strdup(path); /* , 1 + slashpos - path); */
      if(!(*dname)) {
	 perror("memory");
	 exit(1);
      }
      (*dname)[1 + slashpos - path] = 0;
      if (lastdir && !strcmp(lastdir, *dname)) {
	 /***   same as previous directory   ***/
	 free (*dname);
	 *dname = lastdir;
	 return 0;
      }
      else {
	 /***   different directory   ***/
	 if (lastdir)
	   free (lastdir);
	 lastdir = *dname;
	 return 1;
      }
   }
   else {
      /***   no directory specified   ***/
      if (lastdir) {
	 free (lastdir);
	 lastdir = NULL;
      };
      *dname = NULL;
      *fname = (char *)path;
      return 0;
   }
}


static void print_title(void)
{
#ifdef USE_ESD   
   if (param.audio_out) 
     fprintf(stderr,"\n%s (simple MP3 player) version %s with esound\n", PACKAGE,VERSION);
   else
#endif
   fprintf(stderr,"\n%s (simple MP3 player) version %s\n", PACKAGE,VERSION);
}

static void usage(char *d)
{
    FILE *o = stderr;

    fprintf(stderr,"\nusage: %s [option(s)] [file(s) | URL(s) | -]\n", PACKAGE);
    fprintf(stderr,"supported options:\n\n");
    fprintf(o," -t     --test             testmode (no output)\n");
    fprintf(o," -1     --left             Play only left channel\n");
    fprintf(o," -2     --right            Play only right channel\n");
    fprintf(o," -z     --shuffle          Shuffle song-list before playing\n");
    fprintf(o," -Z     --random           full random play\n");
    fprintf(o,"        --force-3dnowex    Force use of 3DNowEx! optimized routine\n");
    fprintf(o,"        --force-3dnow      Force use of 3DNow! optimized routine\n");
    fprintf(o,"        --force-sse        Force use of SSE optimized routine\n");
    fprintf(o,"        --force-mmx        Force use of MMX optimized routine\n");
    fprintf(o,"        --force-fpu        Force use of floating-pointer routine\n");
#ifdef USE_ESD
    fprintf(o,"        --esd              Audio out via esound\n");
    fprintf(o,"        --oss              Audio out via OSS\n");
#endif
    fprintf(o," -v     --version          Prints version and exit\n");
   
    exit(0);
}


