// #define NEWBUFFERING
//#define DEBUG_RESYNC

/* 1 frame = 4608 byte PCM (32*36*2*2) */

#ifdef __GNUC__
#define LOCAL static inline
#else
#define LOCAL static _inline
#endif

//#undef LOCAL
//#define LOCAL

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <math.h>

#include "mp3_config.h"
#include "mp3.h"
#include "mpg123engine.h"
#include "huffman.h"
#include "bswap.h"

#ifdef RUNTIME_CPUDETECT
#include <x86cpucaps.h>
#endif
struct modcaps modcap;

#ifdef USE_FASTMEMCPY
#include "fastmemcpy.h"

void *(*fastmemcpy_func)(void * to, const void * from, size_t len);
void *(*mem2agpcpy_func)(void * to, const void * from, size_t len);

#include "aclib.h"
#endif

#ifdef ARCH_X86
#define CAN_COMPILE_X86_ASM
#endif

#ifdef USE_JCONV
#include "jconv.h"
#endif
   
FILE* mp3_file=NULL;

struct MP3Info_t MP3Info;
static int MP3_resync=0;
int MP3_lockfile=FALSE;

static const long outscale = 32768;
struct mpstr mp;
struct bitstream_info bsi;

unsigned char *pcm_sample;   /* outbuffer address */
int pcm_point = 0;           /* outbuffer offset */

static struct frame fr;

static int tabsel_123[2][3][16] = {
   { {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,},
     {0,32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384,},
     {0,32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320,} },

   { {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,} }
};

static long freqs[9] = { 44100, 48000, 32000, 22050, 24000, 16000 , 11025 , 12000 , 8000 };


static int mp3_gettag(struct MP3Info_t *mi, FILE *fd);
int (*audio_read)(char *buf,int size);

//#include "tabinit.c"

//extern int mplayer_audio_read(char *buf,int size);
// mplayer must use 'mplayer_audio_read'.

static int audio_read_normal(char *buf, int size)
{
//DBGMSG0("audio_read_normal\n");   
  return fread(buf,1,size,mp3_file);
}

LOCAL int mp3_read(char *buf,int size){
//  int len=fread(buf,1,size,mp3_file);
  int len=(audio_read)(buf,size);
  if(len>0) MP3Info.fpos+=len;
//  if(len!=size) MP3Info.eof=1;
  return len;
}

//void mp3_seek(int pos){
//  fseek(mp3_file,pos,SEEK_SET);
//  return (MP3Info.fpos=ftell(mp3_file));
//}
#if 0
int MP3_jump_to_frame(struct frame *fr, int frame)
{  
   if (!mp3_file) return -1;
/*   mpg123_read_frame_init(); */
   fseek(mp3_file, frame * (fr->framesize +4), SEEK_SET);
   if(!read_frame(&fr, &MP3Info))return(0);
   return 0;
}
#endif

LOCAL int stream_head_read(unsigned char *hbuf,unsigned long *newhead){
  if(mp3_read(hbuf,4) != 4) return FALSE;
#if defined(CAN_COMPILE_X86_ASM)
  *newhead = bswap_32(*((unsigned long *)hbuf));
#else
  /*
   * we may not be able to address unaligned 32-bit data on non-x86 cpus.
   * Fall back to some portable code.
   */
  *newhead = 
      hbuf[0] << 24 |
      hbuf[1] << 16 |
      hbuf[2] <<  8 |
      hbuf[3];
#endif
  return TRUE;
}

LOCAL int stream_head_shift(unsigned char *hbuf,unsigned long *head){
  *((unsigned long *)hbuf) >>= 8;
  if(mp3_read(hbuf+3,1) != 1) return 0;
  *head <<= 8;
  *head |= hbuf[3];
  return 1;
}

/*
 * decode a header and write the information
 * into the frame structure
 */
LOCAL int decode_header(struct frame *fr,unsigned long newhead, struct MP3Info_t *mi){

    // head_check:
    if( (newhead & 0xffe00000) != 0xffe00000 ||  
        (newhead & 0x0000fc00) == 0x0000fc00) return FALSE;

    fr->lay = 4-((newhead>>17)&3);
//    if(fr->lay!=3) return FALSE;

    if( newhead & ((long)1<<20) ) {
      fr->lsf = (newhead & ((long)1<<19)) ? 0x0 : 0x1;
      fr->mpeg25 = 0;
    } else {
      fr->lsf = 1;
      fr->mpeg25 = 1;
    }

    if(fr->mpeg25)
      fr->sampling_frequency = 6 + ((newhead>>10)&0x3);
    else
      fr->sampling_frequency = ((newhead>>10)&0x3) + (fr->lsf*3);

    if(fr->sampling_frequency>8) return FALSE;  // valid: 0..8

    fr->error_protection = ((newhead>>16)&0x1)^0x1;
    fr->bitrate_index = ((newhead>>12)&0xf);
    fr->padding   = ((newhead>>9)&0x1);
    fr->extension = ((newhead>>8)&0x1);
    fr->mode      = ((newhead>>6)&0x3);
    fr->mode_ext  = ((newhead>>4)&0x3);
    fr->copyright = ((newhead>>3)&0x1);
    fr->original  = ((newhead>>2)&0x1);
    fr->emphasis  = newhead & 0x3;

    mi->channels = fr->stereo    = (fr->mode == MPG_MD_MONO) ? 1 : 2;

    if(!fr->bitrate_index){
//      fprintf(stderr,"Free format not supported.\n");
      return FALSE;
    }

switch(fr->lay){
  case 2:
    mi->bitrate=tabsel_123[fr->lsf][1][fr->bitrate_index];
    mi->samplerate=freqs[fr->sampling_frequency];
    fr->framesize = (long) mi->bitrate * 144000;
    fr->framesize /= mi->samplerate;
    mi->framesize=fr->framesize;
    fr->framesize += fr->padding - 4;
    break;
  case 3:
    if(fr->lsf)
      mp.ssize = (fr->stereo == 1) ? 9 : 17;
    else
      mp.ssize = (fr->stereo == 1) ? 17 : 32;
    if(fr->error_protection) mp.ssize += 2;

    mi->bitrate=tabsel_123[fr->lsf][2][fr->bitrate_index];
    mi->samplerate=freqs[fr->sampling_frequency];
    fr->framesize  = (long) mi->bitrate * 144000;
    fr->framesize /= mi->samplerate<<(fr->lsf);
    mi->framesize=fr->framesize;
    fr->framesize += fr->padding - 4;
    break;
  case 1:
//    fr->jsbound = (fr->mode == MPG_MD_JOINT_STEREO) ? (fr->mode_ext<<2)+4 : 32;
    mi->bitrate=tabsel_123[fr->lsf][0][fr->bitrate_index];
    mi->samplerate=freqs[fr->sampling_frequency];
    fr->framesize  = (long) mi->bitrate * 12000;
    fr->framesize /= mi->samplerate;
    mi->framesize  = ((fr->framesize+fr->padding)<<2);
    fr->framesize  = mi->framesize-4;
    break;
  default:
    mi->framesize=fr->framesize=0;
//    fprintf(stderr,"Sorry, unsupported layer type.\n");
    return 0;
}
    if(fr->framesize<=0 || fr->framesize>MAXFRAMESIZE) return FALSE;

    return 1;
}

#include "sr2.c"

LOCAL int stream_read_frame_body(int size){

  /* flip/init buffer for Layer 3 */
  mp.bsbufold = mp.bsbuf;
  mp.bsbuf = mp.bsspace[mp.bsnum]+512;
  mp.bsnum = (mp.bsnum + 1) & 1;

  if( mp3_read(mp.bsbuf,size) != size) return 0; // broken frame

  bsi.bitindex = 0;
  bsi.wordpointer = (unsigned char *) mp.bsbuf;
  bsi.bitsleft=8*size;

  return 1;
}


/*****************************************************************
 * read next frame     return number of frames read.
 */
LOCAL int read_frame(struct frame *fr, struct MP3Info_t *mi){
  unsigned long newhead;
  unsigned char hbuf[8];
  int skipped,resyncpos;
  int frames=0;

resync:
  skipped=mi->fpos;
  resyncpos=mi->fpos;

  set_pointer(512);
  mp.fsizeold=fr->framesize;       /* for Layer3 */
  if(!stream_head_read(hbuf,&newhead)) return 0;
  if(!decode_header(fr,newhead,&*mi)){
    // invalid header! try to resync stream!
#ifdef DEBUG_RESYNC
    printf("ReSync: searching for a valid header...  (pos=%X)\n",mi->fpos);
#endif
retry1:
    while(!decode_header(fr,newhead,&*mi)){
      if(!stream_head_shift(hbuf,&newhead)) return 0;
    }
    resyncpos=mi->fpos-4;
    // found valid header
#ifdef DEBUG_RESYNC
    printf("ReSync: found valid hdr at %X  fsize=%ld  ",resyncpos,fr->framesize);
#endif
    if(!stream_read_frame_body(fr->framesize)) return 0; // read body
    set_pointer(512);
    mp.fsizeold=fr->framesize;       /* for Layer3 */
    if(!stream_head_read(hbuf,&newhead)) return 0;
    if(!decode_header(fr,newhead, &*mi)){
      // invalid hdr! go back...
#ifdef DEBUG_RESYNC
      printf("INVALID\n");
#endif
//      mp3_seek(resyncpos+1);
      if(!stream_head_read(hbuf,&newhead)) return 0;
      goto retry1;
    }
#ifdef DEBUG_RESYNC
    printf("OK!\n");
    ++frames;
#endif
  }

  skipped=resyncpos-skipped;
//  if(skipped && !MP3_resync) printf("\r%d bad bytes skipped  (resync at 0x%X)                          \n",skipped,resyncpos);

//  printf("%8X [%08X] %d %d (%d)%s%s\n",MP3_fpos-4,newhead,fr->framesize,fr->mode,fr->mode_ext,fr->error_protection?" CRC":"",fr->padding?" PAD":"");

  /* read main data into memory */
  if(!stream_read_frame_body(fr->framesize)){
    printf("\nBroken frame at 0x%X                                                  \n",resyncpos);
    return 0;
  }
  ++frames;

  if(MP3_resync){
    MP3_resync=0;
    if(frames==1) goto resync;
  }

  return frames;
}

//int _has_mmx = 0;

#include "layer2.c"
#include "layer3.c"
#include "layer1.c"

/******************************************************************************/
/*           PUBLIC FUNCTIONS                  */
/******************************************************************************/

static int tables_done_flag=0;

/* It's hidden from gcc in assembler */
extern void __attribute__ (( __stdcall__ )) dct64_MMX_3dnow(real *, real *, real *);
extern void __attribute__ (( __stdcall__ )) dct64_MMX_3dnowex(real *, real *, real *);
extern void __attribute__ (( __stdcall__ )) dct64_MMX_sse(real *, real *, real *);
extern void __attribute__ (( __stdcall__ )) dct64_MMX(real *, real *, real *);
void __attribute__ (( __stdcall__ )) (*dct64_MMX_func)(real *, real *, real *);

//#include "cpudetect.h"

//#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>

// Init decoder tables.  Call first, once!
#ifdef USE_FAKE_MONO
void MP3_Init(int fakemono){
#else
void MP3_Init(){
#endif
   int accel = 0;

   mp.fsizeold = mp.ssize = mp.bsnum = 0;
   mp.bsbufold=mp.bsspace[0]+512;
   mp.bsbuf=mp.bsspace[1]+512;

   audio_read = audio_read_normal; // set to normal read
   
#ifdef USE_FASTMEMCPY
   fastmemcpy_func = fast_memcpy_nosimd;
   mem2agpcpy_func = mem2agpcpy_nosimd;
#endif

   memset(&modcap,0,sizeof(struct modcaps));
#ifdef RUNTIME_CPUDETECT
   struct simdcaps sdcap;
   memset(&sdcap,0,sizeof(struct simdcaps));
   
   x86cpucaps_simdall(&sdcap,0);
   
    if(sdcap.has_mmx) 	 accel = MODE_MMX;
//    if(sdcap.has_mmxext) accel = MODE_MMX;
#ifdef USE_SSE
    if(sdcap.has_sse) 	accel = MODE_SSE;
#endif
    if(sdcap.has_3dnow) accel = MODE_3DNOW;
    if(sdcap.has_3dnowext) accel = MODE_3DNOWEXT;
   
   modcap.has_mmx=sdcap.has_mmx;
   modcap.has_mmxext=sdcap.has_mmxext;
   modcap.has_sse=sdcap.has_sse;
   modcap.has_sse2=sdcap.has_sse2;
   modcap.has_3dnowext=sdcap.has_3dnowext;
   modcap.has_3dnow=sdcap.has_3dnow;

#else
   accel = MODE_FPU;
   #ifdef HAVE_MMX
   accel = MODE_MMX;
   #endif
   #ifdef HAVE_3DNOW
   accel = MODE_3DNOW;
   #endif
   #ifdef HAVE_3DNOWEXT
   accel = MODE_3DNOWEXT;
   #endif
   #if defined(HAVE_SSE2) || defined(HAVE_SSE)
   accel = MODE_SSE;
   #endif  
#endif

#ifdef USE_FAKE_MONO
   MP3_set_synth_functions(accel,fakemono);
#else
   MP3_set_synth_functions(accel);
#endif   

DBGMSG0("MP3_init done\n");
}


#ifdef USE_FAKE_MONO
void MP3_set_synth_functions(int accel, int fakemono)
#else
void MP3_set_synthfunctions(int accel)
#endif
{
   int family = 0;
   
#ifdef USE_FAKE_MONO
    if (fakemono == 1)
        fr.synth = synth_1to1_l;
    else if (fakemono == 2)
        fr.synth = synth_1to1_r;
    else
        fr.synth = synth_1to1;
#else
    fr.synth = synth_1to1;
#endif
    fr.synth_mono = synth_1to1_mono2stereo;
    dct36_func = dct36;

#ifndef CAN_COMPILE_X86_ASM
    make_decode_tables(outscale);
    synth_func = NULL;
    DBGMSG1(" %s: using generic C decore!\n", MODNAME);
# ifdef USE_FASTMEMCPY
    fastmemcpy_func = fast_memcpy_nosimd;
    mem2agpcpy_func = mem2agpcpy_nosimd;
    DBGMSG1(" %s: using generic memcpy\n", MODNAME);
# endif
   
#else /* CAN_COMPILE_X86_ASM */
    if (accel > MODE_FPU)
    {
//	_has_mmx = 1;
	make_decode_tables_MMX(outscale);
	DBGMSG1(" %s: made decode tables with MMX optimization\n", MODNAME);
    }
    else
     {
//	_has_mmx = 0;
	make_decode_tables(outscale);
	DBGMSG1(" %s: made decode tables with FPU\n", MODNAME);
     }

# ifdef USE_SSE
    if(accel == MODE_SSE)
    {
	/* SSE version is buggy */
	synth_func = synth_1to1_MMX;
	dct64_MMX_func = dct64_MMX_sse;
	DBGMSG1(" %s: using SSE optimized decore!\n", MODNAME);
#  ifdef USE_FASTMEMCPY
        if (modcap.has_sse2 == 1) {
	   fastmemcpy_func = fast_memcpy_SSE;
	   mem2agpcpy_func = mem2agpcpy_SSE;
	   DBGMSG1(" %s: using SSE2 optimized fastmemcpy\n", MODNAME);
        } else {
	   fastmemcpy_func = fast_memcpy_MMX2;
	   mem2agpcpy_func = mem2agpcpy_MMX2;
	   DBGMSG1(" %s: using MMX2 optimized fastmemcpy\n", MODNAME);
        }
#  endif
    }
    else
# endif
    if (accel == MODE_3DNOWEXT)
    {
	synth_func=synth_1to1_MMX;
	dct36_func=dct36_3dnowex;
	dct64_MMX_func=dct64_MMX_3dnowex;
	DBGMSG1(" %s: using 3DNow!Ex optimized decore!\n", MODNAME);
# ifdef USE_FASTMEMCPY
        fastmemcpy_func = fast_memcpy_3DNow;
        mem2agpcpy_func = mem2agpcpy_3DNow;
        DBGMSG1(" %s: using 3DNow! optimized fastmemcpy\n", MODNAME);
# endif
    }
    else
    if (accel == MODE_3DNOW)
    {
	synth_func = synth_1to1_MMX;
	dct36_func = dct36_3dnow;
	dct64_MMX_func = dct64_MMX_3dnow;
	DBGMSG1(" %s: using 3DNow! optimized decore!\n", MODNAME);
# ifdef USE_FASTMEMCPY
        fastmemcpy_func = fast_memcpy_3DNow;
        mem2agpcpy_func = mem2agpcpy_3DNow;
        DBGMSG1(" %s: using 3DNow! optimized fastmemcpy\n", MODNAME);
# endif
    }
    else
    if (accel == MODE_MMX)
    {
        synth_func = synth_1to1_MMX;
	dct64_MMX_func = dct64_MMX;
	DBGMSG1(" %s: using MMX optimized decore!\n", MODNAME);
# ifdef USE_FASTMEMCPY
        fastmemcpy_func = fast_memcpy_MMX;
        mem2agpcpy_func = mem2agpcpy_MMX;
        DBGMSG1(" %s: using MMX optimized fastmemcpy\n", MODNAME);
# endif
    }
    else
# ifdef ENABLE_PENTIUMOPT     
     {
	family=x86cpucaps_cpumodel(GET_FAMILY);
	if (family >= 5)
	  {
	     synth_func = synth_1to1_pent;
	     DBGMSG1(" %s: using Pentium optimized decore!\n", MODNAME);
#  ifdef USE_FASTMEMCPY
	     fastmemcpy_func = fast_memcpy_nosimd;
	     mem2agpcpy_func = mem2agpcpy_nosimd;
           DBGMSG1(" %s: using generic memcpy\n", MODNAME);
#  endif
	  }
    else
# endif
    {
	synth_func = NULL; /* use default c version */
	DBGMSG1(" %s: using generic C decore!\n", MODNAME);
# ifdef USE_FASTMEMCPY
        fastmemcpy_func = fast_memcpy_nosimd;
        mem2agpcpy_func = mem2agpcpy_nosimd;
        DBGMSG1(" %s: using generic memcpy\n", MODNAME);
# endif
    }
# ifdef ENABLE_PENTIUMOPT     
    }
# endif
#endif /* CAN_COMPILE_X86_ASM */

    modcap.decoder = accel;
    fr.down_sample = 0;
    fr.down_sample_sblimit = SBLIMIT>>(fr.down_sample);
    init_layer2();
    init_layer3(fr.down_sample_sblimit);
    tables_done_flag = 1;
    DBGMSG0(" %s: init layer2&3 finished, tables done\n", MODNAME);
}
   
   
//#if 0
void MP3_Close(){
DBGMSG0("mp3close\n");   
  MP3Info.eof = 1;
  if (mp3_file) fclose(mp3_file);
  mp3_file = NULL;
}

// Open a file, init buffers. Call once per file!
int MP3_Open(char *filename,int buffsize){
   DBGMSG0("mp3_open\n");   

   memset(&MP3Info, 0, sizeof(struct MP3Info_t));
   MP3Info.bps = 2;

   MP3Info.eof = 1;   // lock decoding
   MP3Info.pause = 1; // lock playing
   if (mp3_file) MP3_Close(); // close prev. file
   MP3Info.frames = 0;

   mp3_file = fopen(filename,"rb");
//  printf("MP3_Open: file='%s'",filename);
//  if(!mp3_file){ printf(" not found!\n"); return 0;} else printf("Ok!\n");
   if (!mp3_file) return 0;

   mp3_gettag(&MP3Info, mp3_file);
   MP3_GetSongInfo(&MP3Info, filename);
//  fseek(mp3_file,0,SEEK_SET); // moved into PrintTAG

//  MP3_InitBuffers(buffsize);
#ifdef USE_FAKE_MONO
  if(!tables_done_flag) MP3_Init(0);
#else
  if(!tables_done_flag) MP3_Init();
#endif   
  MP3Info.eof = 0;  // allow decoding
  MP3Info.pause = 0; // allow playing

  return MP3Info.filesize;
}

//#endif

// Read & decode a single frame. Called by sound driver.
int MP3_DecodeFrame(unsigned char *hova,short single){
   pcm_sample = hova;
   pcm_point = 0;

   if (!read_frame(&fr, &MP3Info)) return(0);
   if (single == -2) { set_pointer(512); return(1); }
   if (fr.error_protection) getbits(16); /* skip crc */
   fr.single = single;
   switch(fr.lay){
     case 2: do_layer2(&fr,single);break;
     case 3: do_layer3(&fr,single);break;
     case 1: do_layer1(&fr,single);break;
     default:
         return 0;	// unsupported
   }
   ++MP3Info.frames;
   return (pcm_point ? pcm_point : 2);
}

// Prints last frame header in ascii.
void MP3_PrintHeader(){
        static char *modes[4] = { "Stereo", "Joint-Stereo", "Dual-Channel", "Single-Channel" };
        static char *layers[4] = { "???" , "I", "II", "III" };

        printf("\rMPEG %s, Layer %s, %ld Hz %d kbit %s, BPF: %ld\n",
                fr.mpeg25 ? "2.5" : (fr.lsf ? "2.0" : "1.0"),
                layers[fr.lay],freqs[fr.sampling_frequency],
    tabsel_123[fr.lsf][fr.lay-1][fr.bitrate_index],
                modes[fr.mode],(long)compute_bpf(&fr)); // bpf=fr.framesize+4
        printf("Channels: %d, copyright: %s, original: %s, CRC: %s, emphasis: %d\n",
                fr.stereo,fr.copyright?"Yes":"No",
                fr.original?"Yes":"No",fr.error_protection?"Yes":"No",
                fr.emphasis);
}

#if 1
#include "genre.h"

static int mp3_gettag(struct MP3Info_t *mi, FILE *fd)
{
   char id3data[128];
   unsigned char genre;
   int fsize;
#ifdef USE_JCONV
   char tmpstr[31],*cnvstr;
#endif   

DBGMSG0("gettag\n");
   memset(&mi->tag1,0,sizeof(struct id3tag));
   
   fseek(fd,0,SEEK_END);
   fsize = ftell(fd);
   if (fseek(fd, -1*sizeof(id3data), SEEK_END))
     {
	mi->filesize = fsize;
	return 0;
     }
   if (fread(&id3data,1,sizeof(id3data),fd) != sizeof(id3data) || 
      id3data[0]!='T' || id3data[1]!='A' || id3data[2]!='G')
     {
	mi->filesize = fsize;
	return 0;
     }

#ifdef USE_JCONV
   DBGMSG0("jconv\n");
   
        memset(tmpstr, 0, 31);
        memcpy(tmpstr, &id3data[3], 30);
        cnvstr = jconv_strdup_conv_fullauto(tmpstr);
        strncpy(mi->tag1.title,cnvstr,30);
        strip_spaces(mi->tag1.title,30);
        free(cnvstr);

        memset(tmpstr, 0, 31);
        memcpy(tmpstr, &id3data[33], 30);
        cnvstr = jconv_strdup_conv_fullauto(tmpstr);
        strncpy(mi->tag1.artist,cnvstr,30);
//DBGMSG0("artist(prev) %s, artist(conv) %s, %s\n", tmpstr, mi->tag1.artist,cnvstr);
        strip_spaces(mi->tag1.artist,30);
        free(cnvstr);
   
        memset(tmpstr, 0, 31);
        memcpy(tmpstr, &id3data[63], 30);
        cnvstr = jconv_strdup_conv_fullauto(tmpstr);
        strncpy(mi->tag1.album,cnvstr,30);
        strip_spaces(mi->tag1.album,30);
        free(cnvstr);
   
        memcpy(mi->tag1.year, &id3data[93], 4);
   
        memset(tmpstr, 0, 31);
        memcpy(tmpstr, &id3data[97], 30);
        cnvstr = jconv_strdup_conv_fullauto(tmpstr);
        strncpy(mi->tag1.comment,cnvstr,30);
        strip_spaces(mi->tag1.comment,30);
        free(cnvstr);
#else
        memcpy(mi->tag1.title,&id3data[3],30);
        strip_spaces(mi->tag1.title,30);
        memcpy(mi->tag1.artist,&id3data[33],30);
        strip_spaces(mi->tag1.artist,30);
        memcpy(mi->tag1.album,&id3data[63],30);
        strip_spaces(mi->tag1.album,30);
        memcpy(mi->tag1.year,&id3data[93],4);
        memcpy(mi->tag1.comment,&id3data[97],30);
        strip_spaces(mi->tag1.comment,30);
#endif   

        genre = id3data[127];
        if ( genre <= sizeof(genre_table)/sizeof(*genre_table) ) {
                strncpy(mi->tag1.genre, genre_table[genre], 30);
        } else {
                strncpy(mi->tag1.genre,"Unknown",30);
        }
   
   fseek(fd,0,SEEK_SET);

   mi->filesize = fsize - sizeof(id3data);
   return 1;
};
   
int MP3_PrintTAG(void){
DBGMSG0("mp3 printtag\n");   
DBGMSG0("title %s\n",MP3Info.tag1.title);   
   printf("\n");
   printf("Title  : %-30s  Artist: %s\n",MP3Info.tag1.title,MP3Info.tag1.artist);
   printf("Album  : %-30s  Year  : %-4s\n",MP3Info.tag1.album,MP3Info.tag1.year);
   printf("Comment: %-30s  Genre : %s\n",MP3Info.tag1.comment,MP3Info.tag1.genre);
   printf("\n");
   
   return MP3Info.filesize;
}

#endif

