#include "dxhead.h"

static int head_check(unsigned long head)
{
   if ((head & 0xffe00000) != 0xffe00000)
     return FALSE;
   if (!((head >> 17) & 3))
     return FALSE;
   if (((head >> 12) & 0xf) == 0xf)
     return FALSE;
   if (!((head >> 12) & 0xf))
     return FALSE;
   if (((head >> 10) & 0x3) == 0x3)
     return FALSE;
   if (((head >> 19) & 1) == 1 && ((head >> 17) & 3) == 3 && ((head >> 16) & 1) == 1)
     return FALSE;
   if ((head & 0xffff0000) == 0xfffe0000)
     return FALSE;
   
   return TRUE;
}

static unsigned int convert_to_header(unsigned char * buf)
{
   return (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
}

static double compute_tpf(struct frame *fr)
{
   const int bs[4] = {0, 384, 1152, 1152};
   double tpf;

   tpf = bs[fr->lay];
   tpf /= freqs[fr->sampling_frequency] << (fr->lsf);
   return tpf;
}

static double compute_bpf(struct frame *fr)
{
   double bpf;

   switch (fr->lay)
     {
//      case 1:
//	bpf = tabsel_123[fr->lsf][0][fr->bitrate_index];
//	bpf *= 12000.0 * 4.0;
//	bpf /= freqs[fr->sampling_frequency] << (fr->lsf);
//	break;
      case 2:
      case 3:
	bpf = tabsel_123[fr->lsf][fr->lay - 1][fr->bitrate_index];
	bpf *= 144000;
	bpf /= freqs[fr->sampling_frequency] << (fr->lsf);
	break;
      default:
	bpf = 1.0;
     }
   
   return bpf;
}

int calc_numframes(struct frame *fr)
{
   return (int) (MP3Info.filesize / compute_bpf(fr));
}

static unsigned int strip_spaces(char *src, size_t n)
/* strips trailing spaces from string of length n
   returns length of adjusted string */
{
   char *space = NULL,	/* last space in src */
   *start = src;

   while (n--)
     switch (*src++)
       {
	case '\0':
	  n = 0;	/* breaks out of while loop */
	  src--;
	  break;
	case ' ':
	  if (space == NULL)
	    space = src - 1;
	  break;
	default:
	  space = NULL;	/* don't terminate intermediate spaces */
	  break;
       }
   if (space != NULL)
     {
	src = space;
	*src = '\0';
     }
   return src - start;
}

// return value = filesize
// update MP3Info.tag,length,
int MP3_GetSongInfo(struct MP3Info_t *mi, char *filename)
{
   unsigned int head;
   unsigned char tmp[4], *buf;
   xing_header_t xing_header;
   double tpf, bpf;
   unsigned int len;
   FILE *fd;

   memset(&*mi,0,sizeof(struct MP3Info_t));
   fd = fopen(filename,"rb");
   if (!fd) return 0;
   
// get id3tag(v1)
   mi->has_tag = mp3_gettag(&*mi,fd);
//DBGMSG0("mp3_getlength\n");
//DBGMSG0("mp3_getlength: numframes=%d\n", calc_numframes(&fr));
   
   fseek(fd, 0, SEEK_SET);
   if (fread(tmp, 1, 4, fd) != 4) 
     return 0;
   
   head = convert_to_header(tmp);
   while (!head_check(head))
     {
	head <<= 8;
	if (fread(tmp, 1, 1, fd) != 1)
	  return 0;
	head |= tmp[0];
     }
   if (decode_header(&fr, head, &*mi))
     {
	buf = malloc(mi->framesize + 4);
	fseek(fd, -4, SEEK_CUR);
	fread(buf, 1, mi->framesize + 4, fd);
	tpf = compute_tpf(&fr);
	fr.have_xing_header = get_xing_header(&xing_header, buf);
	if (fr.have_xing_header)
	  {
	     free(buf);
	     mi->length=((int) (tpf * xing_header.frames * 1000));
//DBGMSG0("mp3_getlength1=%d\n",mi->length);
	     return mi->length;
	  }
	free(buf);
	bpf = compute_bpf(&fr);
	fseek(fd, 0, SEEK_END);
	len = ftell(fd);
	fseek(fd, -128, SEEK_END);
	fread(tmp, 1, 3, fd);
	if (!strncmp(tmp, "TAG", 3))
	  len -= 128;
	mi->length=((int) ((int)(len / bpf) * tpf * 1000));
DBGMSG0("mp3_getlength2=%d\n",mi->length);
	return mi->length;
     }

   return 0;
}

