
static GtkWidget *mpg123alt_configurewin = NULL;
static GtkWidget *decoder_vbox, *decoder_frame, *decoder_vbox, 
  *auto_select, *decoder_sse, *decoder_3dnowex, *decoder_3dnow, 
  *decoder_mmx, *decoder_fpu;

static void init_decoder_selection(void);


static void select_decoder(int m)
{
DBGMSG0("select_decoder : %d\n", m);
   switch (m) {
    case MODE_SSE:
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(decoder_sse), TRUE);
      break;
    case MODE_3DNOWEXT:
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(decoder_3dnowex), TRUE);
      break;
    case MODE_3DNOW:
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(decoder_3dnow), TRUE);
      break;
    case MODE_MMX:
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(decoder_mmx), TRUE);
      break;
    case MODE_FPU:
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(decoder_fpu), TRUE);
      break;
    default:
DBGMSG0("select_decoder : others\n", m);
      break;
      
   }
}

static void auto_select_cb(GtkWidget * w, gpointer data)
{
   if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(auto_select)) == TRUE)
     param.simd_select = MODE_AUTO;
   else
     param.simd_select = param.auto_simd;
   
DBGMSG0("auto\n");
   init_decoder_selection();
}

static void decoder_sse_cb(GtkWidget * w, gpointer data)
{
DBGMSG0("SSE\n");
}

static void decoder_3dnowex_cb(GtkWidget * w, gpointer data)
{
DBGMSG0("3DNow!ex\n");
}

static void decoder_3dnow_cb(GtkWidget * w, gpointer data)
{
DBGMSG0("3DNow!\n");
}

static void decoder_mmx_cb(GtkWidget * w, gpointer data)
{
DBGMSG0("MMX\n");
}

static void decoder_fpu_cb(GtkWidget * w, gpointer data)
{
DBGMSG0("FPU\n");
}

static void set_decoders_selection(int *m, int *s)
{
DBGMSG0("init decoder: auto_simd=%d, simd_select=%d\n", param.auto_simd, param.simd_select);
   if (param.simd_select == MODE_AUTO) {
      *m = param.auto_simd;
      *s = FALSE;
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(auto_select), TRUE);
   } else {
      *m = param.simd_select;
      *s = TRUE;
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(auto_select), FALSE);
   }
}

static void init_decoder_selection(void)
{
   int m,s;
   
   set_decoders_selection(&m, &s);

   gtk_widget_set_sensitive(decoder_sse, s);
   gtk_widget_set_sensitive(decoder_3dnowex, s);
   gtk_widget_set_sensitive(decoder_3dnow, s);
   gtk_widget_set_sensitive(decoder_mmx, s);
   gtk_widget_set_sensitive(decoder_fpu, s);
   
DBGMSG0("init_decoder: %d\n",m);
   select_decoder(m);
}


static void mpg123alt_configurewin_ok(GtkWidget * widget, gpointer data)
{
   ConfigFile *cfg;
   gchar *filename;
   
   if (GTK_TOGGLE_BUTTON(auto_select)->active)
     param.simd_select = MODE_AUTO;
   else {
     if (GTK_TOGGLE_BUTTON(decoder_sse)->active)
       param.simd_select = MODE_SSE;
     if (GTK_TOGGLE_BUTTON(decoder_3dnowex)->active)
       param.simd_select = MODE_3DNOWEXT;
     if (GTK_TOGGLE_BUTTON(decoder_3dnow)->active)
       param.simd_select = MODE_3DNOW;
     if (GTK_TOGGLE_BUTTON(decoder_mmx)->active)
       param.simd_select = MODE_MMX;
     if (GTK_TOGGLE_BUTTON(decoder_fpu)->active)
       param.simd_select = MODE_FPU;
   }
   
#ifdef USE_FAKE_MONO
   if (param.simd_select != MODE_AUTO) MP3_set_synth_functions(param.simd_select, param.force_mono);
#else
   if (param.simd_select != MODE_AUTO) MP3_set_synth_functions(param.simd_select);
#endif

   filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
   cfg = xmms_cfg_open_file(filename);
   if (!cfg) cfg = xmms_cfg_new();
   
   xmms_cfg_write_int(cfg, "MPG123alt", "simd_decoder", param.simd_select);
   
   xmms_cfg_write_file(cfg, filename);
   xmms_cfg_free(cfg);
   g_free(filename);
   gtk_widget_destroy(mpg123alt_configurewin);
}

void mpg123alt_configure(void)
{
   GtkWidget *vbox1;
   GtkWidget *decoders_frame;
   GtkWidget *vbox2;
   GtkWidget *decoders;
   GSList *decoders_group = NULL;
   GtkWidget *bbox;
   GtkWidget *ok;
   GtkWidget *cancel;

   mpg123alt_configurewin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
   gtk_object_set_data (GTK_OBJECT (mpg123alt_configurewin), 
	       "mpg123alt_configurewin", mpg123alt_configurewin);
   gtk_window_set_title (GTK_WINDOW (mpg123alt_configurewin), 
			 "mpg123alt_configurewin");

   vbox1 = gtk_vbox_new (FALSE, 0);
   gtk_widget_ref (vbox1);
   gtk_object_set_data_full (GTK_OBJECT (mpg123alt_configurewin), "vbox1", 
	 vbox1, (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (vbox1);
   gtk_container_add (GTK_CONTAINER (mpg123alt_configurewin), vbox1);

   decoders_frame = gtk_frame_new (NULL);
   gtk_widget_ref (decoders_frame);
   gtk_object_set_data_full (GTK_OBJECT (mpg123alt_configurewin), 
       "decoders_frame", decoders_frame, (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (decoders_frame);
   gtk_box_pack_start (GTK_BOX (vbox1), decoders_frame, TRUE, TRUE, 0);
   gtk_container_set_border_width (GTK_CONTAINER (decoders_frame), 1);

   vbox2 = gtk_vbox_new (FALSE, 0);
   gtk_widget_ref (vbox2);
   gtk_object_set_data_full (GTK_OBJECT (mpg123alt_configurewin), 
	"vbox2", vbox2, (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (vbox2);
   gtk_container_add (GTK_CONTAINER (decoders_frame), vbox2);

   auto_select = gtk_check_button_new_with_label ("Enable Automatic detection");
   gtk_widget_ref (auto_select);
   gtk_object_set_data_full (GTK_OBJECT (mpg123alt_configurewin), 
	 "auto_select", auto_select, (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (auto_select);
   gtk_box_pack_start (GTK_BOX (vbox2), auto_select, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(auto_select), "clicked",
		      GTK_SIGNAL_FUNC(auto_select_cb), NULL);
   
   decoders = gtk_vbox_new (FALSE, 0);
   gtk_widget_ref (decoders);
   gtk_object_set_data_full (GTK_OBJECT (mpg123alt_configurewin), 
       "decoders", decoders, (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (decoders);
   gtk_box_pack_start (GTK_BOX (vbox2), decoders, TRUE, TRUE, 0);

   decoder_sse = gtk_radio_button_new_with_label (decoders_group, "use SSE optimized decoder");
   decoders_group = gtk_radio_button_group (GTK_RADIO_BUTTON (decoder_sse));
   gtk_widget_ref (decoder_sse);
   gtk_object_set_data_full (GTK_OBJECT (mpg123alt_configurewin), 
       "decoder_sse", decoder_sse, (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (decoder_sse);
   gtk_box_pack_start (GTK_BOX (decoders), decoder_sse, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(decoder_sse), "clicked",
		      GTK_SIGNAL_FUNC(decoder_sse_cb), NULL);

   decoder_3dnowex = gtk_radio_button_new_with_label (decoders_group, 
			     "use 3DNow!ex optimized decoder");
   decoders_group = gtk_radio_button_group (GTK_RADIO_BUTTON (decoder_3dnowex));
   gtk_widget_ref (decoder_3dnowex);
   gtk_object_set_data_full (GTK_OBJECT (mpg123alt_configurewin), 
      "decoder_3dnowex", decoder_3dnowex, (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (decoder_3dnowex);
   gtk_box_pack_start (GTK_BOX (decoders), decoder_3dnowex, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(decoder_3dnowex), "clicked",
		      GTK_SIGNAL_FUNC(decoder_3dnowex_cb), NULL);

   decoder_3dnow = gtk_radio_button_new_with_label (decoders_group, 
				   "use 3DNow! optimized decoder");
   decoders_group = gtk_radio_button_group (GTK_RADIO_BUTTON (decoder_3dnow));
   gtk_widget_ref (decoder_3dnow);
   gtk_object_set_data_full (GTK_OBJECT (mpg123alt_configurewin), 
       "decoder_3dnow", decoder_3dnow, (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (decoder_3dnow);
   gtk_box_pack_start (GTK_BOX (decoders), decoder_3dnow, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(decoder_3dnow), "clicked",
		      GTK_SIGNAL_FUNC(decoder_3dnow_cb), NULL);

   decoder_mmx = gtk_radio_button_new_with_label (decoders_group, "use MMX optimized decoder");
   decoders_group = gtk_radio_button_group (GTK_RADIO_BUTTON (decoder_mmx));
   gtk_widget_ref (decoder_mmx);
   gtk_object_set_data_full (GTK_OBJECT (mpg123alt_configurewin), 
	 "decoder_mmx", decoder_mmx, (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (decoder_mmx);
   gtk_box_pack_start (GTK_BOX (decoders), decoder_mmx, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(decoder_mmx), "clicked",
		      GTK_SIGNAL_FUNC(decoder_mmx_cb), NULL);

   decoder_fpu = gtk_radio_button_new_with_label (decoders_group, 
						  "use generic (FPU) decoder");
   decoders_group = gtk_radio_button_group (GTK_RADIO_BUTTON (decoder_fpu));
   gtk_widget_ref (decoder_fpu);
   gtk_object_set_data_full (GTK_OBJECT (mpg123alt_configurewin),
	"decoder_fpu", decoder_fpu, (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (decoder_fpu);
   gtk_box_pack_start (GTK_BOX (decoders), decoder_fpu, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(decoder_fpu), "clicked",
		     GTK_SIGNAL_FUNC(decoder_fpu_cb), NULL);

   bbox = gtk_hbutton_box_new ();
   gtk_widget_ref (bbox);
   gtk_object_set_data_full (GTK_OBJECT (mpg123alt_configurewin), "bbox", bbox,
                            (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (bbox);
   gtk_box_pack_start (GTK_BOX (vbox1), bbox, TRUE, TRUE, 0);

   init_decoder_selection();
   
   ok = gtk_button_new_with_label("Ok");
   gtk_signal_connect(GTK_OBJECT(ok), "clicked", GTK_SIGNAL_FUNC(mpg123alt_configurewin_ok), NULL);
   GTK_WIDGET_SET_FLAGS(ok, GTK_CAN_DEFAULT);
   gtk_box_pack_start(GTK_BOX(bbox), ok, TRUE, TRUE, 0);
   gtk_widget_grab_default(ok);
   gtk_widget_show (ok);
   
   cancel = gtk_button_new_with_label("Cancel");
   gtk_signal_connect_object(GTK_OBJECT(cancel), "clicked", 
      GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(mpg123alt_configurewin));
   GTK_WIDGET_SET_FLAGS(cancel, GTK_CAN_DEFAULT);
   gtk_box_pack_start(GTK_BOX(bbox), cancel, TRUE, TRUE, 0);
   gtk_widget_show (cancel);
   
   gtk_widget_show_all(mpg123alt_configurewin);
   
}

