/*  mpg123alt - An XMMS input plugin using libMP3
**
**  Copyright (C) 2002 Osamu Kayasono
**
**  This program is free software ; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation ; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY ; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program ; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>

//#ifdef ENABLE_NLS
//#include <locale.h>
//#endif

#include <xmms/plugin.h>
#include <xmms/configfile.h>
#include <xmms/titlestring.h>
#include <xmms/util.h>

#include "../mp3_config.h"
#include "../mpg123engine.h"
#include "../mp3.h"
#include "mpg123alt.h"

#include "../mp3dec/player.h"

#include "configure.c"

#define BUFFLEN 16384
static unsigned char buff[BUFFLEN];

#define NAME		"MPG123 Alternative input plugin"
#define AUTHOR		"Osamu Kayasono"
#define CONTACT		"jacobi@jcom.home.ne.jp"
#define HOMEPAGE	"http://members.jcom.home.ne.jp/jacobi/xmms.html"

static GtkWidget *dialog = NULL, *button = NULL, *label = NULL ;

static char description [] = NAME " " VERSION ;
static char title_str [] = "About " NAME " " VERSION ;
static char about_str [] =
  "Copyright (C) 2000 " AUTHOR "\n"
  "Written and maintained by " AUTHOR " <" CONTACT ">.\n"
  "Homepage: " HOMEPAGE ".\n\n"
  "This program is free software ; you can redistribute it and/or modify \n"
  "it under the terms of the GNU General Public License as published by \n"
  "the Free Software Foundation ; either version 2 of the License, or \n"
  "(at your option) any later version. \n \n"
  "This program is distributed in the hope that it will be useful, \n"
  "but WITHOUT ANY WARRANTY ; without even the implied warranty of \n"
  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  \n"
  "See the GNU General Public License for more details. \n\n"
  "You should have received a copy of the GNU General Public \n"
  "License along with this program ; if not, write to \n"
  "the Free Software Foundation, Inc., \n"
  "59 Temple Place, Suite 330, \n"
  "Boston, MA  02111-1307  USA" ;


#define	TITLE_LEN 128
static gchar song_title[TITLE_LEN];
static int decoding ;
static int pause_now ;
static int seek_time = -1 ;
static int single=0;

static struct AudioInfo_s AudioInfo;

static pthread_t decode_thread ;

static	void 	plugin_init (void) ;
static	void	about_dialog (void) ;
static	int	is_our_file (char *filename) ;
static	void 	play_start (char *filename) ;
static	void 	play_stop (void) ;
static	void 	do_pause (short pause) ;
//static	void 	file_seek (int time) ;
static	int	get_time (void) ;
static	void 	get_song_info (char *filename, char **title, int *length) ;

struct parameter param;

static
InputPlugin mpg123alt_ip =
{	NULL,				/* Filled in by xmms	*/
	NULL,				/* Filled in by xmms	*/
	description,			/* The description that is shown in the preferences box 	*/
	plugin_init,			/* Called when the plugin is loaded	*/
	about_dialog,			/* Show the about_dialog box	*/
	mpg123alt_configure, 		/* Show the configure box	*/
	is_our_file,			/* Return 1 if the plugin can handle the file	*/
	NULL,				/* Scan dir	*/
	play_start,			/* Play file	*/
	play_stop,			/* Stop	*/
	do_pause,			/* Pause	*/
	NULL,				/* Seek	*/
	NULL,				/* Set the equalizer, most plugins won't be able to do this	*/
	get_time,			/* Get the time, usually returns the output plugins output time	*/
	NULL,				/* Get volume	*/
	NULL,				/* Set volume	*/
	NULL,				/* OBSOLETE, DO NOT USE!	*/
	NULL,				/* OBSOLETE, DO NOT USE!	*/
	NULL,				/* Send data to the visualization plugins	*/
	NULL,				/* Fill in the stuff that is shown in the player window	*/
	NULL,				/* Show some text in the song title box. Filled in by xmms	*/
	get_song_info,			/* Function to grab the title string	*/
	NULL,				/* Bring up an info window for the filename passed in	*/
	NULL				/* Handle to the current output plugin. Filled in by xmms	*/
} ; /* mpg123alt_ip */


static gchar *format_song_title(gchar *filename, struct MP3Info_t *mi);
static gchar *extname(const char *filename);

static gchar *format_song_title(gchar *filename, struct MP3Info_t *mi)
{
DBGMSG0("format_song_title\n");

   gchar *ret = NULL;
   TitleInput *input;
   
   XMMS_NEW_TITLEINPUT(input);

   MP3_GetSongInfo(&*mi, filename);
   
   if (mi->has_tag == 1) 
     {
	input->performer = g_strstrip(mi->tag1.artist);
	input->album_name = g_strstrip(mi->tag1.album);
	input->track_name = mi->tag1.title;
	input->year = atoi(mi->tag1.year);
	input->genre = mi->tag1.genre;
	input->comment = g_strstrip(mi->tag1.comment);
     }
   input->file_name = g_basename(filename);
   input->file_path = filename;
   input->file_ext = extname(filename);
   
   ret = xmms_get_titlestring(xmms_get_gentitle_format(), input);
   g_free(input);
   
   if (!ret)
     {
	ret = g_strdup(g_basename(filename));
	if (extname(ret) != NULL) *(extname(ret) -1 )= '\0';
     }
   
   return ret;
}


static char *extname(const char *filename)
{
   gchar *ext = strrchr(filename, '.');

   if (ext != NULL)
     ++ext;

   return ext;
}

static	
void plugin_init (void) 
{
   ConfigFile *cfg;
   
DBGMSG0("plugin_init\n");
   memset(&param, 0, sizeof(struct parameter));

   decoding = FALSE ;
   pause_now = FALSE ;
   dialog = NULL ;
   seek_time = -1 ;
   
   cfg = xmms_cfg_open_default_file();
   
   xmms_cfg_read_int(cfg, "MPG123alt", "simd_decoder", &param.simd_select);
   xmms_cfg_free(cfg);
   
// MPEG Audio init:
#ifdef USE_FAKE_MONO
   MP3_Init(0);
   param.auto_simd = modcap.decoder;
   if (param.simd_select != MODE_AUTO) MP3_set_synth_functions(param.simd_select, param.force_mono);
#else
   MP3_Init();
   param.auto_simd = modcap.decoder;
   if (param.simd_select != MODE_AUTO) MP3_set_synth_functions(param.simd_select);
#endif

} /* plugin_int */

static	
void about_dialog (void) 
{
   if (dialog) 
     return ;

   dialog = gtk_dialog_new () ;
   gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
		       GTK_SIGNAL_FUNC (gtk_widget_destroyed), &dialog) ;
   gtk_window_set_title (GTK_WINDOW (dialog), title_str) ;
   gtk_window_set_policy (GTK_WINDOW (dialog), FALSE, FALSE, FALSE) ;
   gtk_container_border_width (GTK_CONTAINER (dialog), 5) ;
   
   label = gtk_label_new (about_str) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), label, TRUE, TRUE, 0) ;
   gtk_widget_show (label) ;
   
   button = gtk_button_new_with_label (" Close ") ;
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
	       GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog)) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), button,
		     FALSE, FALSE, 0) ;
  
   gtk_widget_show (button) ;
   gtk_widget_show (dialog) ;
   gtk_widget_grab_focus (button) ;
} /* about_dialog */

static	
int is_our_file (char *filename) 
{
DBGMSG0("our_file\n");
   int r;
   struct MP3Info_t mi;
   
   r = MP3_GetSongInfo(&mi,filename);
   if (r == 0 ) 
     return FALSE ;
	
DBGMSG0("our_file=true\n");
   return TRUE ;
}

static	
void *play_loop (void *arg) 
{//	short buffer [BUFFLEN] ;
DBGMSG0("play loop\n");
   int k;
   int len = 0;
	
   decoding = TRUE ;
//   len = MP3_DecodeFrame(buff, single);
   
   while (decoding)
     {	
	if (pause_now == FALSE){
	   len = MP3_DecodeFrame(buff, single);
//DBGMSG0("rate=%d, channels=%d, bitrate=%d, \n", MP3Info.samplerate, MP3Info.channels, MP3Info.bitrate);   
	   mpg123alt_ip.set_info (song_title, AudioInfo.length, MP3Info.bitrate*1000, MP3Info.samplerate, MP3Info.channels);
	}
	   
	if (len >= 0 &&  decoding) {
	   mpg123alt_ip.output->write_audio (buff, len * sizeof(unsigned char)) ;
	   mpg123alt_ip.add_vis_pcm (mpg123alt_ip.output->written_time (), HOST_SIGNED_SHORT, MP3Info.channels, len * sizeof(unsigned char), buff) ;
	   
	   /* Not sure why this is needed, but it can stutter at
	    ** the start of the first file if this isn't here.
	    */
	   xmms_usleep (100) ; 
	}
	else 
	  xmms_usleep (100000) ;
	
	/* Loop while waiting for the output buffer to empty. */
	while ((mpg123alt_ip.output->buffer_free () < (len * sizeof(unsigned char))) && decoding)
	  xmms_usleep (10000) ;
	
//		/* Do seek if seek_time is valid. */
//		if (seek_time > 0)
//		{	sf_seek (sndfile, seek_time * sfinfo.samplerate, SEEK_SET) ;
//			mpg123alt_ip.output->flush (seek_time * 1000) ;
//			seek_time = -1 ;
  // 			} ;
//
     } ; /* while (decoding) */

   pthread_exit (NULL) ;
} /* play_loop */

static	
void play_start (char *filename) 
{	
DBGMSG0("play start\n");
   struct MP3Info_t mi;

   strncpy(song_title, format_song_title(filename, &mi), TITLE_LEN);
   AudioInfo.length = mi.length;
   if (AudioInfo.length == 0)
		return ;
   
   MP3_Open(filename, BUFFLEN);
   if (MP3Info.channels == 2)
     single = -1;
   else
     single = 3;
   
   AudioInfo.start_time = GetTimer();
//DBGMSG0("rate=%d, channels=%d, bitrate=%d, \n", MP3Info.samplerate, MP3Info.channels, MP3Info.bitrate);   
   if (! mpg123alt_ip.output->open_audio (HOST_SIGNED_SHORT, MP3Info.samplerate, MP3Info.channels))
     {	MP3_Close();
	return ;
     } ;

   mpg123alt_ip.set_info (song_title, AudioInfo.length, MP3Info.bitrate*1000, MP3Info.samplerate, MP3Info.channels) ;

   pthread_create (&decode_thread, NULL, play_loop, NULL) ;
  
   while (decoding == FALSE)
     xmms_usleep (1000) ;

} /* play_start */

static	
void play_stop (void) 
{	
   decoding = FALSE ;
   pause_now = FALSE ;
	
   pthread_join (decode_thread, NULL) ;
   mpg123alt_ip.output->close_audio () ;

   MP3_Close();

   seek_time = -1 ;
} /* play_stop */

static	
void do_pause (short p) 
{	
   if (p) 
     pause_now = TRUE ;
   else
     pause_now = FALSE ;

   mpg123alt_ip.output->pause (p) ;
} /* do_pause */

//static	
//void file_seek (int time) 
//{	
//	if (! sfinfo.seekable)
//		return ;
//
//	seek_time = time ;
//
//	while (seek_time != -1)
//		xmms_usleep (10000) ;
//} /* file_seek */

static	
int get_time (void)
{
   if ( ! (mpg123alt_ip.output->buffer_playing () && decoding))
     return -1 ;
		
   return mpg123alt_ip.output->output_time () ;
} /* get_time */

static 
void get_song_info (char *filename, char **title, int *length) 
{
   struct MP3Info_t mi;
   (*length) = -1;
   (*title) = NULL;
   
   
   (*title) = format_song_title(filename, &mi);
//   MP3_GetSongInfo(&mi, filename);
//   if (mi.has_tag == 1) 
//     (*title) = g_strdup(mi.tag1.title);
//   else
//     (*title) = g_strdup(g_basename(filename));
   (*length) = mi.length;
   
DBGMSG0("songinfo hastag=%d, %s\n",mi.has_tag, (*title));
}

InputPlugin *get_iplugin_info (void) 
{	return &mpg123alt_ip ;
} /* get_iplugin_info */



